@@ -738,6 +738,13 @@ LogicProgram& LogicProgram::addRule(const Rule& rule) {
 			}
 		}
 	}
+	else if (statsId_ == 0) {
+		for (Potassco::AtomSpan::iterator it = Potassco::begin(rule.head), end = Potassco::end(rule.head); it != end; ++it) {
+			if (!validAtom(*it) || (isNew(*it) && getAtom(*it)->supports() == 0)) {
+				auxData_->skippedHeads.insert(*it);
+			}
+		}
+	}
 	rule_.clear();
 	return *this;
 }
@@ -1040,6 +1047,7 @@ void LogicProgram::prepareExternals() {
 			*j++ = encodeExternal(id, value);
 		}
 	}
+	external.erase(j, external.end());
 	for (VarVec::const_iterator it = external.begin(), end = external.end(); it != end; ++it) {
 		atomState_.clearRule(decodeExternal(*it).first);
 	}
@@ -1143,6 +1151,7 @@ void LogicProgram::prepareProgram(bool checkSccs) {
 	finalizeDisjunctions(p, sccs);
 	prepareComponents();
 	prepareOutputTable();
+	freezeAssumptions();
 	if (incData_ && options().distTrue) {
 		for (Var a = startAtom(), end = startAuxAtom(); a != end; ++a) {
 			if (isSentinel(getRootAtom(a)->literal())) {
@@ -1163,11 +1172,12 @@ void LogicProgram::prepareProgram(bool checkSccs) {
 }
 void LogicProgram::freezeTheory() {
 	if (theory_) {
+		const IdSet& skippedHeads = auxData_->skippedHeads;
 		for (TheoryData::atom_iterator it = theory_->currBegin(), end = theory_->end(); it != end; ++it) {
 			const Potassco::TheoryAtom& a = **it;
 			if (isFact(a.atom()) || !isNew(a.atom())) { continue; }
 			PrgAtom* atom = resize(a.atom());
-			if (!atom->frozen() && atom->supports() == 0 && atom->relevant()) {
+			if (!atom->frozen() && atom->supports() == 0 && atom->relevant() && skippedHeads.count(a.atom()) == 0) {
 				pushFrozen(atom, value_free);
 			}
 		}
@@ -1407,6 +1417,17 @@ void LogicProgram::prepareOutputTable() {
 		}
 	}
 }
+
+// Make assumptions/externals exempt from sat-preprocessing
+void LogicProgram::freezeAssumptions() {
+	for (VarVec::const_iterator it = frozen_.begin(), end = frozen_.end(); it != end; ++it) {
+		ctx()->setFrozen(getRootAtom(*it)->var(), true);
+	}
+	for (Potassco::LitVec::const_iterator it = auxData_->assume.begin(), end = auxData_->assume.end(); it != end; ++it) {
+		ctx()->setFrozen(getLiteral(Potassco::id(*it)).var(), true);
+	}
+}
+
 // add (completion) nogoods
 bool LogicProgram::addConstraints() {
 	ClauseCreator gc(ctx()->master());
@@ -1440,7 +1461,7 @@ bool LogicProgram::addConstraints() {
 		if (!toConstraint(getAtom(*it), *this, gc)) { return false; }
 	}
 	// add atoms from this step
-	const bool freezeAll = incData_ && ctx()->satPrepro.get() != 0;
+	const bool freezeAll = incData_ != 0;
 	const uint32 hiAtom  = startAuxAtom();
 	uint32 id = startAtom();
 	for (AtomList::const_iterator it = atoms_.begin()+startAtom(), end = atoms_.end(); it != end; ++it, ++id) {
@@ -1667,7 +1688,9 @@ bool LogicProgram::simplifySum(Head_t ht, const Potassco::AtomSpan& head, const
 				meta.hash += hashLit(p);
 			}
 			else { // Merge duplicate lits
-				w = (std::find_if(out.wlits_begin(), out.wlits_end(), IsLit(Potassco::lit(*it)))->weight += w);
+				Potassco::WeightLit_t* pos = std::find_if(out.wlits_begin(), out.wlits_end(), IsLit(toInt(p)));
+				POTASSCO_ASSERT(pos != out.wlits_end());
+				w = (pos->weight += w);
 				++dirty;
 			}
 			if (w > maxW) { maxW = w; }
