@@ -62,9 +62,9 @@ class Controller:
         while True:
             signal.signal(signal.SIGINT, pyInt)
             try:
-                try: input = raw_input
-                except NameError: pass
-                line = input('> ')
+                try: input_ = raw_input
+                except NameError: input_ = input
+                line = input_('> ')
                 signal.signal(signal.SIGINT, signal.SIG_IGN)
             except EOFError:
                 signal.signal(signal.SIGINT, signal.SIG_IGN)
@@ -129,7 +129,7 @@ class SolveThread(Thread):
     def run(self):
         while True:
             if self.state == SolveThread.STATE_SOLVE:
-                f = self.prg.solve(on_model=self.on_model, on_finish=self.on_finish, async=True)
+                f = self.prg.solve(on_model=self.on_model, on_finish=self.on_finish, async_=True)
             msg = self.input.receive()
             if self.state == SolveThread.STATE_SOLVE:
                 f.cancel()
